# VBoxGtk: A VirtualBox GTK+ GUI
# Copyright (C) 2008-2011 Francisco J. Vazquez Araujo, Spain
# franjva at gmail dot com

# This file is part of VBoxGtk.

# VBoxGtk is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# VBoxGtk is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with VBoxGtk.  If not, see <http://www.gnu.org/licenses/>.


"""
Adapter for the VirtualBox API.
"""


import os
import threading
import vboxapi

import iface
import util


class WatcherThread(threading.Thread):
    """Watches a VM session in a separate thread and notifies the
    interface of the vm state changes"""

    def __init__(self, vm):
        threading.Thread.__init__(self)
        self.vm = vm

    def run(self):
        session = get_session(self.vm)
        console = session.console
        evsource = console.eventSource
        listener = evsource.createListener()
        evs = [cts.VBoxEventType_OnStateChanged]
        evsource.registerListener(listener, evs, False)
        keep_running = True
        while keep_running:
            try:
                event = evsource.getEvent(listener, 1000)
            except:
                keep_running = False
            else:
                if event is not None:
                    iface.vm_updated(self.vm, True)
                if self.vm.state < cts.MachineState_Running:
                    keep_running = False
        evsource.unregisterListener(listener)
        close_session(self.vm)  # make sure the session is removed
        return 0


class SysProps():
    """Struct to store the important system parameters"""
    def __init__(self, min_ncpus, max_ncpus, min_ram, max_ram, min_vram,
                 max_vram, max_hdsize, vm_folder):
        self.min_ncpus = min_ncpus
        self.max_ncpus = max_ncpus
        self.min_ram = min_ram
        self.max_ram = max_ram
        self.min_vram = min_vram
        self.max_vram = max_vram
        self.max_hdsize = max_hdsize
        self.vm_folder = vm_folder


# Singleton

def modmethod(func):
    class modproxy(object):
        __getattribute__ = func.func_globals.__getitem__
        __setattr__ = func.func_globals.__setitem__
    self = modproxy()
    def call_with_method(*args,**kwargs):
        return func(self,*args,**kwargs)
    call_with_method.func_name = func.func_name
    return call_with_method

@modmethod
def init(self):
    """Reads the needed info from VirtualBoxManager"""
    self.vbox_mgr = vboxapi.VirtualBoxManager(None, None)
    self.vbox = vbox_mgr.vbox
    self.mgr = vbox_mgr.mgr
    self.cts = vbox_mgr.constants
    self.sysprops = SysProps(vbox.systemProperties.minGuestCPUCount,
                             8, # FIXME: vbox reports 32 CPUs
                             vbox.systemProperties.minGuestRAM,
                             4096, # FIXME: vbox reports 2TB max RAM
                             vbox.systemProperties.minGuestVRAM,
                             vbox.systemProperties.maxGuestVRAM,
                             vbox.systemProperties.infoVDSize,
                             vbox.systemProperties.defaultMachineFolder)
    self.os_types = vbox.getGuestOSTypes()
    self.dvd_images = vbox.getDVDImages()
    self.dvd_host_drives = vbox.host.getDVDDrives()
    self.sessions = {}
    self.vms = vbox.getMachines()
    for vm in self.vms:
        if vm.state >= cts.MachineState_Running:
            watch_vm(vm)

def exit():
    """Must make sure to close all the pending sessions before exiting"""
    for vm in vms:
        close_session(vm)
    return False

## Session management

def get_session(vm=None):
    """If no vm is passed, we create a new session. Otherwise we get the
    session from vm.lockMachine."""
    if vm is None:
        session = mgr.getSessionObject(vbox)
    elif vm.id in sessions:
        session = sessions[vm.id]
    else:
        session = get_session()
        vm.lockMachine(session, cts.LockType_Shared)
        sessions[vm.id] = session
    return session

def close_session(vm):
    if not vm.id in sessions:
        return
    if sessions[vm.id].state == cts.SessionState_Locked:
        sessions[vm.id].unlockMachine()
    del sessions[vm.id]

## VM

def watch_vm(vm):
    thread = WatcherThread(vm);
    thread.daemon = True
    thread.start()

def create_vm(vm_name, os_type, hd_size, hd_dynamic):
    try:
        vm = vbox.createMachine('', vm_name, os_type, None, False)
        vm.addStorageController('IDE Controller', cts.StorageBus_IDE)
        ot = vbox.getGuestOSType(os_type)
        vm.memorySize = ot.recommendedRAM
        vm.VRAMSize = ot.recommendedVRAM
        vbox.registerMachine(vm)
        vms.append(vm)
        vm_name_full = vbox.composeMachineFilename(vm_name, None)
        vm_dir = os.path.dirname(vm_name_full)
        hd_path = os.path.join(vm_dir, '%s.img' % (vm_name,))
        hd = vbox.createHardDisk('', hd_path)
        mtype = (cts.MediumVariant_Standard if hd_dynamic
                 else cts.MediumVariant_Fixed)
        progress = hd.createBaseStorage(hd_size * 1024 * 1024, mtype)
        progress.waitForCompletion(-1)
        if hd.state == cts.MediumState_NotCreated:
            delete_vm(-1, False)
            raise Exception(_("Could not create hard disk "
                              "(Make sure the file name is not in use)"))
        attach_device(-1, 'hd', hd.id)
        attach_device(-1, 'dvd', None)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def delete_vm(vm_number, delete_media=True):
    vm = vms[vm_number]
    try:
        close_session(vm)
        media = vm.unregister(cts.CleanupMode_DetachAllReturnHardDisksOnly)
        if delete_media:
            progress = vm.delete(media)
            progress.waitForCompletion(-1)
        del vms[vm_number]
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def start_vm(vm_number):
    vm = vms[vm_number]
    close_session(vm)
    session = get_session()
    progress = vm.launchVMProcess(session, 'sdl', None)
    progress.waitForCompletion(-1)
    msg1 = _("The VM could not be started")
    if session.state == cts.SessionState_Locked:
        session.unlockMachine()
    else:
        msg2 = _("This usually means that the vboxdrv module is not "
                 "loaded. Try again after loading the module.")
        return (1, "%s (%s)" % (msg1, msg2))
    if vm.state < cts.MachineState_Running:
        msg2 = _("Make sure all the VirtualBox modules are loaded and the "
                 "VM configuration is correct.")
        return (1, "%s (%s)" % (msg1, msg2))
    iface.vm_updated(vm, False)
    watch_vm(vm)
    return (0, None)

def stop_vm(vm_number):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        console = session.console
        console.powerDown()
        close_session(vm)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def suspend_vm(vm_number):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        console = session.console
        console.saveState()
        close_session(vm)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def update_settings(vm_number, ncpus, hwvirtext, mem, vmem):
    vm = vms[vm_number]
    session = get_session(vm)
    session.machine.CPUCount = ncpus
    session.machine.setHWVirtExProperty(cts.HWVirtExPropertyType_Enabled,
                                        hwvirtext)
    session.machine.memorySize = mem
    session.machine.VRAMSize = vmem
    session.machine.saveSettings()

def get_medium_info(vm_number, mtype):
    vm = vms[vm_number]
    port = 0 if mtype == 'hd' else 1
    sc = vm.getStorageControllers()[0].name
    medatt = vm.getMediumAttachment(sc, port, 0)
    medium = medatt.medium
    passthrough = medatt.passthrough
    medium_id = str(medium.id) if medium else None
    return (medium_id, passthrough)

def attach_device(vm_number, devtype, dev_id, passthrough=False):
    if devtype == 'hd':
        port = 0
        dev_type = cts.DeviceType_HardDisk
        try:
            device = vbox.findMedium(dev_id, dev_type)
        except Exception as e:
            return (1, str(e))
    else:
        port = 1
        dev_type = cts.DeviceType_DVD
        device = util.find(dvd_host_drives, 'id', dev_id)
    vm = vms[vm_number]
    session = get_session(vm)
    sc = session.machine.getStorageControllers()[0].name
    try:
        session.machine.detachDevice(sc, port, 0)
    except:  # no device was connected
        pass
    try:
        session.machine.attachDevice(sc, port, 0, dev_type, device)
        if devtype == 'dvd':
            session.machine.passthroughDevice(sc, port, 0, passthrough)
        session.machine.saveSettings()
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def register_dvd(path):
    try:
        devtype = cts.DeviceType_DVD
        mode = cts.AccessMode_ReadOnly
        dvd = vbox.openMedium(path, devtype, mode, False)
        dvd_images.append(dvd)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def mount_dvd(vm_number, dvd_id):
    vm = vms[vm_number]
    try:
        dvd = vbox.findMedium(dvd_id, cts.DeviceType_DVD) if dvd_id else None
    except Exception as e:
        return (1, str(e))
    session = get_session(vm)
    sc = session.machine.getStorageControllers()[0].name
    try:
        session.machine.mountMedium(sc, 1, 0, dvd, False)
        session.machine.saveSettings()
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def remove_dvd(dvd_id):
    dvd = util.find(dvd_images, 'id', dvd_id)
    try:
        dvd.close()
        dvd_images.remove(dvd)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def update_devices(vm_number, attachment, device, intnet, hostif,
                   mac, cable, audiodev):
    vm = vms[vm_number]
    session = get_session(vm)
    net = session.machine.getNetworkAdapter(0)
    net.adapterType = device
    net.internalNetwork = intnet
    net.bridgedInterface = hostif
    net.MACAddress = mac
    net.cableConnected = cable
    net.attachmentType = attachment
    if audiodev == -1:
        session.machine.audioAdapter.enabled = False
    else:
        session.machine.audioAdapter.enabled = True
        session.machine.audioAdapter.audioDriver = audiodev
    session.machine.saveSettings()

def create_shared(vm_number, name, folder, writable, automount):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        session.machine.createSharedFolder(name, folder, writable, automount)
        session.machine.saveSettings()
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def remove_shared(vm_number, shared_name):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        session.machine.removeSharedFolder(shared_name)
        session.machine.saveSettings()
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def get_snapshot_tree(vm_number):
    vm = vms[vm_number]
    try:
        root = vm.findSnapshot(None)
    except Exception as e:
        return []
    tree = [(root, None)]
    def traverse_tree(root):
        for child in root.getChildren():
            tree.append((child, root.id))
            traverse_tree(child)
    traverse_tree(root)
    return tree

def take_snapshot(vm_number, name, description = ''):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        console = session.console
        progress = console.takeSnapshot(name, description)
        progress.waitForCompletion(-1)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def edit_snapshot(vm_number, uuid, name):
    vm = vms[vm_number]
    try:
        snapshot = vm.findSnapshot(uuid)
        snapshot.name = name
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def delete_snapshot(vm_number, uuid):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        console = session.console
        progress = console.deleteSnapshot(uuid)
        progress.waitForCompletion(-1)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

def restore_snapshot(vm_number, uuid):
    vm = vms[vm_number]
    try:
        session = get_session(vm)
        console = session.console
        snapshot = vm.findSnapshot(uuid)
        progress = console.restoreSnapshot(snapshot)
        progress.waitForCompletion(-1)
    except Exception as e:
        return (1, str(e))
    else:
        return (0, None)

